namespace Syntriax.Engine.Core;

/// <summary>
/// Provides extension methods for <see cref="Vector3D"/> type.
/// </summary>
public static class Vector3DExtensions
{
    /// <summary>
    /// Returns the <see cref="Vector2D"/> representation of the <see cref="Vector3D"/>.
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The <see cref="Vector2D"/> representation of the provided <see cref="Vector3D"/>.</returns>
    public static Vector2D As2D(this Vector3D vector) => new(vector.X, vector.Y);

    /// <summary>
    /// Calculates the length of the <see cref="Vector3D"/>.
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The length of the <see cref="Vector3D"/>.</returns>
    public static float Length(this Vector3D vector) => Vector3D.Length(vector);

    /// <summary>
    /// Calculates the squared length of the <see cref="Vector3D"/>.
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The squared length of the <see cref="Vector3D"/>.</returns>
    public static float LengthSquared(this Vector3D vector) => Vector3D.LengthSquared(vector);

    /// <summary>
    /// Calculates the distance between two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="from">The starting <see cref="Vector3D"/>.</param>
    /// <param name="to">The ending <see cref="Vector3D"/>.</param>
    /// <returns>The distance between the two <see cref="Vector3D"/>s.</returns>
    public static float Distance(this Vector3D from, Vector3D to) => Vector3D.Distance(from, to);

    /// <summary>
    /// Returns the <see cref="Vector3D"/> with its components inverted.
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The inverted <see cref="Vector3D"/>.</returns>
    public static Vector3D Invert(this Vector3D vector) => Vector3D.Invert(vector);

    /// <summary>
    /// Adds two <see cref="Vector3D"/>s component-wise.
    /// </summary>
    /// <param name="vector">The first <see cref="Vector3D"/>.</param>
    /// <param name="vectorToAdd">The vector <see cref="Vector3D"/> to be added.</param>
    /// <returns>The result of the addition.</returns>
    public static Vector3D Add(this Vector3D vector, Vector3D vectorToAdd) => Vector3D.Add(vector, vectorToAdd);

    /// <summary>
    /// Subtracts one <see cref="Vector3D"/> from another component-wise.
    /// </summary>
    /// <param name="vector">The first <see cref="Vector3D"/>.</param>
    /// <param name="vectorToSubtract">The <see cref="Vector3D"/> to be subtracted.</param>
    /// <returns>The result of the subtraction.</returns>
    public static Vector3D Subtract(this Vector3D vector, Vector3D vectorToSubtract) => Vector3D.Subtract(vector, vectorToSubtract);

    /// <summary>
    /// Multiplies a <see cref="Vector3D"/> by a scalar value.
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to multiply.</param>
    /// <param name="value">The scalar value to multiply with.</param>
    /// <returns>The result of the multiplication.</returns>
    public static Vector3D Multiply(this Vector3D vector, float value) => Vector3D.Multiply(vector, value);

    /// <summary>
    /// Divides a <see cref="Vector3D"/> by a scalar value.
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to divide.</param>
    /// <param name="value">The scalar value to divide with.</param>
    /// <returns>The result of the division.</returns>
    public static Vector3D Divide(this Vector3D vector, float value) => Vector3D.Divide(vector, value);

    /// <summary>
    /// Returns a <see cref="Vector3D"/> with the absolute values of each component.
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The <see cref="Vector3D"/> with absolute values.</returns>
    public static Vector3D Abs(this Vector3D vector) => Vector3D.Abs(vector);

    /// <summary>
    /// Reflects a <see cref="Vector3D"/> off a surface with the specified normal.
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to reflect.</param>
    /// <param name="normal">The normal <see cref="Vector3D"/> of the reflecting surface.</param>
    /// <returns>The reflected <see cref="Vector3D"/>.</returns>
    public static Vector3D Reflect(this Vector3D vector, Vector3D normal) => Vector3D.Reflect(vector, normal);

    /// <summary>
    /// Normalizes the <see cref="Vector3D"/> (creates a <see cref="Vector3D"/> with the same direction but with a length of 1).
    /// </summary>
    /// <param name="vector">The input <see cref="Vector3D"/>.</param>
    /// <returns>The normalized <see cref="Vector3D"/>.</returns>
    public static Vector3D Normalize(this Vector3D vector) => Vector3D.Normalize(vector);

    /// <summary>
    /// Creates a <see cref="Vector3D"/> pointing from one point to another.
    /// </summary>
    /// <param name="from">The starting point.</param>
    /// <param name="to">The ending point.</param>
    /// <returns>The <see cref="Vector3D"/> pointing from <paramref name="from"/> to <paramref name="to"/>.</returns>
    public static Vector3D FromTo(this Vector3D from, Vector3D to) => Vector3D.FromTo(from, to);

    /// <summary>
    /// Scales a <see cref="Vector3D"/> by another <see cref="Vector3D"/> component-wise.
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to scale.</param>
    /// <param name="scale">The <see cref="Vector3D"/> containing the scaling factors for each component.</param>
    /// <returns>The scaled <see cref="Vector3D"/>.</returns>
    public static Vector3D Scale(this Vector3D vector, Vector3D scale) => Vector3D.Scale(vector, scale);

    /// <summary>
    /// Rotates a <see cref="Vector3D"/> by the specified angle (in radians).
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to rotate.</param>
    /// <param name="normal">The <see cref="Vector3D"/> to rotate around.</param>
    /// <param name="angleInRadian">The angle to rotate by, in radians.</param>
    /// <returns>The rotated <see cref="Vector3D"/>.</returns>
    public static Vector3D Rotate(this Vector3D vector, Vector3D normal, float angleInRadian) => Vector3D.Rotate(vector, normal, angleInRadian);

    /// <summary>
    /// Returns the component-wise minimum of two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <returns>The <see cref="Vector3D"/> containing the minimum components from both input <see cref="Vector3D"/>s.</returns>
    public static Vector3D Min(this Vector3D left, Vector3D right) => Vector3D.Min(left, right);

    /// <summary>
    /// Returns the component-wise maximum of two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <returns>The <see cref="Vector3D"/> containing the maximum components from both input <see cref="Vector3D"/>s.</returns>
    public static Vector3D Max(this Vector3D left, Vector3D right) => Vector3D.Max(left, right);

    /// <summary>
    /// Clamps each component of a <see cref="Vector3D"/> between the corresponding component of two other <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="vector">The <see cref="Vector3D"/> to clamp.</param>
    /// <param name="min">The <see cref="Vector3D"/> representing the minimum values for each component.</param>
    /// <param name="max">The <see cref="Vector3D"/> representing the maximum values for each component.</param>
    /// <returns>The clamped <see cref="Vector3D"/>.</returns>
    public static Vector3D Clamp(this Vector3D vector, Vector3D min, Vector3D max) => Vector3D.Clamp(vector, min, max);

    /// <summary>
    /// Linearly interpolates between two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="from">The start <see cref="Vector3D"/>.</param>
    /// <param name="to">The end <see cref="Vector3D"/>.</param>
    /// <param name="t">The interpolation parameter (between 0 and 1).</param>
    /// <returns>The interpolated <see cref="Vector3D"/>.</returns>
    public static Vector3D Lerp(this Vector3D from, Vector3D to, float t) => Vector3D.Lerp(from, to, t);

    /// <summary>
    /// Calculates the cross product of two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <returns>The cross product of the two <see cref="Vector3D"/>s.</returns>
    public static Vector3D Cross(this Vector3D left, Vector3D right) => Vector3D.Cross(left, right);

    /// <summary>
    /// Calculates the angle in radians between two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <returns>The angle between the two <see cref="Vector3D"/>s in radians.</returns>
    public static float AngleBetween(this Vector3D left, Vector3D right) => Vector3D.Angle(left, right);

    /// <summary>
    /// Calculates the dot product of two <see cref="Vector3D"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <returns>The dot product of the two <see cref="Vector3D"/>s.</returns>
    public static float Dot(this Vector3D left, Vector3D right) => Vector3D.Dot(left, right);

    /// <summary>
    /// Checks whether two <see cref="Vector3D"/>s are approximately equal within a certain epsilon range.
    /// </summary>
    /// <param name="left">The first <see cref="Vector3D"/>.</param>
    /// <param name="right">The second <see cref="Vector3D"/>.</param>
    /// <param name="epsilon">The maximum difference allowed between components.</param>
    /// <returns>True if the <see cref="Vector3D"/>s are approximately equal, false otherwise.</returns>
    public static bool ApproximatelyEquals(this Vector3D left, Vector3D right, float epsilon = float.Epsilon) => Vector3D.ApproximatelyEquals(left, right, epsilon);
}
